from typing import Union

from lamin_logger import logger
from lnhub_rest.core.account._signup_signin import sign_in_hub, sign_up_hub

from ._settings import settings
from .dev._db import upsert
from .dev._settings_load import load_or_create_user_settings, load_user_settings
from .dev._settings_save import save_user_settings
from .dev._settings_store import user_settings_file_email, user_settings_file_handle


def signup(email: str) -> Union[str, None]:
    """Sign up user."""
    response = sign_up_hub(email)
    if response == "handle-exists":  # handle already exists
        logger.error("The handle already exists. Please choose a different one.")
        return "handle-exists"
    if response == "user-exists":  # user already exists
        logger.error("User already exists! Please login instead: `lamin login`.")
        return "user-exists"
    user_settings = load_or_create_user_settings()
    user_settings.email = email
    save_user_settings(user_settings)
    user_settings.password = response
    save_user_settings(user_settings)
    return None  # user needs to confirm email now


def load_user(email: str = None, handle: str = None) -> Union[str, None]:
    if email is not None:
        settings_file = user_settings_file_email(email)
    if handle is not None:
        settings_file = user_settings_file_handle(handle)
    if settings_file.exists():
        user_settings = load_user_settings(settings_file)
        save_user_settings(user_settings)  # needed to save to current_user.env
        assert user_settings.email is not None
    else:
        user_settings = load_or_create_user_settings()
        if email is None:
            raise RuntimeError(
                "Use your email for your first login in a compute environment. "
                "After that, you can use your handle."
            )
        user_settings.email = email
        user_settings.handle = handle
        save_user_settings(user_settings)

    from ._settings import settings

    settings._user_settings = None  # this is to refresh a settings instance

    return None


def login(
    user: str,
    *,
    password: Union[str, None] = None,
) -> Union[str, None]:
    """Log in user."""
    if "@" in user:
        email, handle = user, None
    else:
        email, handle = None, user
    load_user(email, handle)

    user_settings = load_or_create_user_settings()

    if password:
        user_settings.password = password

    if user_settings.email is None:
        raise RuntimeError("No stored user email, please call: lamin login {user}")

    if user_settings.password is None:
        raise RuntimeError(
            "No stored user password, please call: lamin login <your-email>"
            " --password <your-password>"
        )

    response = sign_in_hub(
        user_settings.email, user_settings.password, user_settings.handle
    )
    if response == "could-not-login":
        return response
    elif response == "complete-signup":
        return response
    else:
        user_id, user_handle, user_name, access_token = response
    if handle is None:
        logger.success(f"Logged in with handle {user_handle} and id {user_id}")
    else:
        logger.success(f"Logged in with email {user_settings.email} and id {user_id}")
    user_settings.id = user_id
    user_settings.handle = user_handle
    user_settings.name = user_name
    user_settings.access_token = access_token
    save_user_settings(user_settings)

    settings._user_settings = None

    # register login of user in instance db
    if settings._instance_exists:
        if not settings.instance._is_db_reachable():
            logger.info("Consider closing the instance: `lamin close`")
            return None
        upsert.user(
            settings.user.email,
            settings.user.id,
            settings.user.handle,
            settings.user.name,
        )

    return None



"""v0.34a1.

Reverts migration 0.15.0.
"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

revision = "c3f38ffe9e03"
down_revision = "6a73c00555b4"


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        op.rename_table(old_table_name="core.user", new_table_name="lnschema_core_user")
        op.rename_table(old_table_name="core.project", new_table_name="lnschema_core_project")
        op.rename_table(old_table_name="core.storage", new_table_name="lnschema_core_storage")
        op.rename_table(old_table_name="core.transform", new_table_name="lnschema_core_transform")
        op.rename_table(old_table_name="core.run", new_table_name="lnschema_core_run")
        op.rename_table(old_table_name="core.file", new_table_name="lnschema_core_file")
        op.rename_table(old_table_name="core.folder", new_table_name="lnschema_core_folder")
        op.rename_table(old_table_name="core.features", new_table_name="lnschema_core_features")
        op.rename_table(old_table_name="core.run_in", new_table_name="lnschema_core_runinput")
        op.rename_table(old_table_name="core.file_features", new_table_name="lnschema_core_filefeatures")
        op.rename_table(old_table_name="core.project_folder", new_table_name="lnschema_core_projectfolder")
        op.rename_table(old_table_name="core.folder_file", new_table_name="lnschema_core_folderfile")
    else:
        op.rename_table(old_table_name="user", new_table_name="lnschema_core_user", schema="core")
        op.rename_table(old_table_name="project", new_table_name="lnschema_core_project", schema="core")
        op.rename_table(old_table_name="storage", new_table_name="lnschema_core_storage", schema="core")
        op.rename_table(old_table_name="transform", new_table_name="lnschema_core_transform", schema="core")
        op.rename_table(old_table_name="run", new_table_name="lnschema_core_run", schema="core")
        op.rename_table(old_table_name="file", new_table_name="lnschema_core_file", schema="core")
        op.rename_table(old_table_name="folder", new_table_name="lnschema_core_folder", schema="core")
        op.rename_table(old_table_name="features", new_table_name="lnschema_core_features", schema="core")
        op.rename_table(old_table_name="run_in", new_table_name="lnschema_core_runinput", schema="core")
        op.rename_table(old_table_name="file_features", new_table_name="lnschema_core_filefeatures", schema="core")
        op.rename_table(old_table_name="project_folder", new_table_name="lnschema_core_projectfolder", schema="core")
        op.rename_table(old_table_name="folder_file", new_table_name="lnschema_core_folderfile", schema="core")
        # there seems to be a bug in alembic autogenerate that doesn't pick this up
        op.execute("alter table core.lnschema_core_user set schema public")
        op.execute("alter table core.lnschema_core_project set schema public")
        op.execute("alter table core.lnschema_core_storage set schema public")
        op.execute("alter table core.lnschema_core_transform set schema public")
        op.execute("alter table core.lnschema_core_run set schema public")
        op.execute("alter table core.lnschema_core_file set schema public")
        op.execute("alter table core.lnschema_core_folder set schema public")
        op.execute("alter table core.lnschema_core_features set schema public")
        op.execute("alter table core.lnschema_core_runinput set schema public")
        op.execute("alter table core.lnschema_core_filefeatures set schema public")
        op.execute("alter table core.lnschema_core_projectfolder set schema public")
        op.execute("alter table core.lnschema_core_folderfile set schema public")


def downgrade() -> None:
    pass



from enum import Enum
from pathlib import Path
from typing import Callable, TypeVar

import anndata as ad
import numpy as np
import pandas as pd
from lndb.dev.upath import UPath
from sqlalchemy.orm.attributes import InstrumentedAttribute

PathLike = TypeVar("PathLike", str, Path, UPath)
DataLike = TypeVar("DataLike", ad.AnnData, pd.DataFrame)
ListLike = TypeVar("ListLike", pd.Series, list, np.array)
SQLModelField = TypeVar("SQLModelField", Callable, InstrumentedAttribute)


class Usage(str, Enum):
    """Data access types."""

    ingest = "ingest"
    insert = "insert"
    select = "select"
    update = "update"
    delete = "delete"
    load = "load"
    link = "link"


class TransformType(Enum):
    pipeline = "pipeline"
    notebook = "notebook"
    app = "app"

    def __repr__(self):
        return self.name

    @classmethod
    def choices(cls):
        return [(item.value, item.name) for item in cls]



from sqlmodel import Field


def current_user_id() -> str:
    from lndb import settings

    return settings.user.id


def current_user_id_as_int() -> int:
    from lndb import settings

    return settings.user.get_id_as_int()


CreatedBy = Field(default_factory=current_user_id, foreign_key="lnschema_core_user.id", index=True)



"""Data lineage (`yvzi`)."""
_schema_id = "yvzi"
_name = "core"
_migration = "1c49c9f9a982"
__version__ = "0.34.0"

from lndb._check_instance_setup import check_instance_setup as _check_instance_setup

import lndb as _lndb

try:
    from lndb import _USE_DJANGO
except Exception:
    _USE_DJANGO = False

_INSTANCE_SETUP = _check_instance_setup()

if _INSTANCE_SETUP:
    from . import dev, link, types

    if not _USE_DJANGO:
        from ._core import (  # type: ignore
            Features,
            File,
            Folder,
            Project,
            Run,
            Storage,
            Transform,
            User,
        )
    else:
        from .models import (  # type: ignore
            Features,
            File,
            Folder,
            Project,
            Run,
            Storage,
            Transform,
            User,
        )



import sys
from pathlib import Path

import nox
from laminci import move_built_docs_to_docs_slash_project_slug, upload_docs_artifact
from laminci.nox import build_docs, login_testuser1, run_pytest  # noqa

nox.options.default_venv_backend = "none"


@nox.session
def lint(session: nox.Session) -> None:
    session.run(*"pip install pre-commit".split())
    session.run("pre-commit", "install")
    session.run("pre-commit", "run", "--all-files")


@nox.session
def install(session: nox.Session) -> None:
    session.run(*"pip install --no-deps .[django]".split())
    session.run(*"git clone https://github.com/laminlabs/lamindb --depth 1".split())
    if sys.platform.startswith("linux"):  # remove version pin when running on CI
        session.run(*"sed -i /lnschema_core/d ./lamindb/pyproject.toml".split())
    session.run(*"pip install ./lamindb[aws,test]".split())


@nox.session()
def build(session: nox.Session) -> None:
    login_testuser1(session)
    run_pytest(session)
    prefix = "." if Path("./lndocs").exists() else ".."
    session.run(*f"pip install {prefix}/lndocs".split())
    session.run(*"lamin init --storage ./docsbuild".split())
    session.run("lndocs")
    upload_docs_artifact()
    move_built_docs_to_docs_slash_project_slug()



"""Types.

.. autosummary::
   :toctree: .

   PathLike
   DataLike
   ListLike
   SQLModelField
   TransformType
"""
from ._types import DataLike, ListLike, PathLike, SQLModelField, TransformType  # noqa



from pathlib import Path, PurePosixPath
from typing import Any, List, Optional, Union

from django.db import models
from django.db.models import Model as BaseORM
from lamin_logger import logger
from nbproject._is_run_from_ipython import is_run_from_ipython
from upath import UPath

from ._users import current_user_id_as_int
from .types import DataLike, ListLike, PathLike, TransformType


class RunInput(models.Model):
    run = models.ForeignKey("Run", on_delete=models.CASCADE)
    file = models.ForeignKey("File", on_delete=models.CASCADE)

    class Meta:
        managed = True


class User(BaseORM):
    email = models.CharField(max_length=64, unique=True)
    handle = models.CharField(max_length=64, unique=True)
    name = models.CharField(max_length=64, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        managed = True


class Storage(BaseORM):
    root = models.CharField(max_length=255)
    type = models.CharField(max_length=63, blank=True, null=True)
    region = models.CharField(max_length=63, blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(User, models.DO_NOTHING, blank=True, null=True, default=current_user_id_as_int)

    class Meta:
        managed = True


class Project(BaseORM):
    name = models.CharField(max_length=64)
    folders = models.ManyToManyField("Folder")
    files = models.ManyToManyField("File")
    created_by = models.ForeignKey(User, models.DO_NOTHING, default=current_user_id_as_int)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        managed = True


class Transform(models.Model):
    name = models.CharField(max_length=63)
    version = models.CharField(max_length=63)
    type = models.CharField(max_length=63, choices=TransformType.choices(), db_index=True, default=(TransformType.notebook if is_run_from_ipython else TransformType.pipeline))
    title = models.CharField(max_length=63, blank=True, null=True)
    reference = models.CharField(max_length=63, blank=True, null=True)
    created_by = models.ForeignKey(User, models.DO_NOTHING)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        managed = True
        unique_together = (("name", "version"),)


class Run(models.Model):
    name = models.CharField(max_length=255, blank=True, null=True)
    external_id = models.CharField(max_length=255, blank=True, null=True)
    transform = models.ForeignKey(Transform, models.DO_NOTHING)
    inputs = models.ManyToManyField("File", through=RunInput, related_name="input_of")
    # outputs on File
    created_by = models.ForeignKey(User, models.DO_NOTHING)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        managed = True

    def __init__(  # type: ignore
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        load_latest: bool = False,
        external_id: Optional[str] = None,
        transform: Optional[Transform] = None,
        inputs: List["File"] = None,
        outputs: List["File"] = None,
    ):
        kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        import lamindb as ln

        global_context = False
        if transform is None:
            if ln.context.transform is not None:
                global_context = True
                transform = ln.context.transform
            else:
                raise ValueError("Either call `ln.Run(transform=transform)` or `ln.track(transform=...)`.")

        if not isinstance(transform, Transform):
            raise TypeError("transform needs to be of type Transform")

        run = None
        if load_latest:
            run = (
                ln.select(
                    ln.Run,
                    transform_id=transform.id,
                    transform_version=transform.version,
                )
                .order_by(ln.Run.created_at.desc())
                .first()
            )
            if run is not None:
                logger.info(f"Loaded: {run}")
        elif id is not None:
            run = ln.select(ln.Run, id=id).one_or_none()
            if run is None:
                raise NotImplementedError("You can currently only pass existing ids")

        if run is None:
            kwargs.update(dict(transform_id=transform.id, transform_version=transform.version))
            super().__init__(**kwargs)
            self._ln_identity_key = None
        else:
            super().__init__(**run.dict())
            self._ln_identity_key = run.id  # simulate query result

        if global_context:
            if run is None:
                added_self = ln.add(self)
                self._ln_identity_key = added_self.id
                logger.success(f"Added: {self}")
            ln.context.run = self


class Features(models.Model):
    id = models.CharField(max_length=63, primary_key=True)
    type = models.CharField(max_length=63)
    created_by = models.ForeignKey(User, models.DO_NOTHING, default=current_user_id_as_int)
    created_at = models.DateTimeField(auto_now=True)
    files = models.ManyToManyField("File")

    class Meta:
        managed = True

    def __init__(  # type: ignore
        self,
        iterable: ListLike = None,
        field: models.CharField = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
        **map_kwargs,
    ):
        kwargs = locals()

        # needed for erroring when passing pd.index
        if kwargs["data"] is not None:
            kwargs.pop("data")
        if kwargs["iterable"] is not None:
            kwargs.pop("iterable")

        kwargs = {k: v for k, v in kwargs.items() if v and k != "self"}
        super().__init__(**kwargs)

    def __new__(
        cls,
        iterable: ListLike = None,
        field: models.CharField = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
        **map_kwargs,
    ):
        if iterable is not None:
            from lamindb._file import get_features_from_data

            features = get_features_from_data(
                iterable=iterable,
                field=field,
                **map_kwargs,
            )
        else:
            features = super().__new__(cls)
        return features


class Folder(models.Model):
    name = models.CharField(max_length=255)
    key = models.CharField(max_length=255, blank=True, null=True)
    storage = models.ForeignKey(Storage, models.DO_NOTHING)
    files = models.ManyToManyField("File")
    created_by = models.ForeignKey(User, models.DO_NOTHING, default=current_user_id_as_int)
    created_at = models.DateTimeField()
    updated_at = models.DateTimeField(blank=True, null=True)

    class Meta:
        managed = True
        unique_together = (("storage", "key"),)

    @property
    def __name__(cls) -> str:
        return "Folder"

    def path(self) -> Union[Path, UPath]:
        """Path on storage."""
        return filepath_from_file_or_folder(self)

    def tree(
        self,
        level: int = -1,
        limit_to_directories: bool = False,
        length_limit: int = 1000,
    ) -> None:
        """Print a visual tree structure."""
        from lamindb._folder import tree

        return tree(
            dir_path=self.path(),
            level=level,
            limit_to_directories=limit_to_directories,
            length_limit=length_limit,
        )

    def __init__(  # type: ignore
        self,
        path: Optional[Union[Path, UPath, str]] = None,
        *,
        # continue with fields
        id: Optional[str] = None,
        name: Optional[str] = None,
        key: Optional[str] = None,
        storage_id: Optional[str] = None,
        files: List["File"] = [],
    ):
        if path is not None:
            from lamindb._folder import get_folder_kwargs_from_data

            kwargs, privates = get_folder_kwargs_from_data(
                path=path,
                name=name,
                key=key,
            )
            if id is not None:
                kwargs["id"] = id
        else:
            kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        super().__init__(**kwargs)
        if path is not None:
            self._local_filepath = privates["local_filepath"]
            self._cloud_filepath = privates["cloud_filepath"]


class File(models.Model):
    name = models.CharField(max_length=255, blank=True, null=True)
    suffix = models.CharField(max_length=63, blank=True, null=True)
    size = models.BigIntegerField(blank=True, null=True)
    hash = models.CharField(max_length=63, blank=True, null=True)
    key = models.CharField(max_length=255, blank=True, null=True)
    run = models.ForeignKey(Run, models.DO_NOTHING, blank=True, null=True, related_name="outputs")
    transform = models.ForeignKey(Transform, models.DO_NOTHING, blank=True, null=True)
    storage = models.ForeignKey(Storage, models.DO_NOTHING)
    # folders from Folders.files
    # features from Features.files
    # input_of from Run.inputs
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(User, models.DO_NOTHING, default=current_user_id_as_int)

    class Meta:
        managed = True
        unique_together = (("storage", "key"),)

    def path(self) -> Union[Path, UPath]:
        """Path on storage."""
        return filepath_from_file_or_folder(self)

    # likely needs an arg `key`
    def replace(
        self,
        data: Union[PathLike, DataLike],
        run: Optional[Run] = None,
        format: Optional[str] = None,
    ) -> None:
        """Replace data object."""
        from lamindb._file import get_file_kwargs_from_data

        if isinstance(data, (Path, str)):
            name_to_pass = None
        else:
            name_to_pass = self.name

        kwargs, privates = get_file_kwargs_from_data(
            data=data,
            name=name_to_pass,
            run=run,
            format=format,
        )

        if kwargs["name"] != self.name:
            logger.warning(f"Your new filename '{kwargs['name']}' does not match the previous filename '{self.name}': to update the name, set file.name = '{kwargs['name']}'")

        if self.key is not None:
            key_path = PurePosixPath(self.key)
            if isinstance(data, (Path, str)):
                new_name = kwargs["name"]  # use the name from the data filepath
            else:
                # do not change the key stem to file.name
                new_name = key_path.stem  # use the stem of the key for in-memory data
            if PurePosixPath(new_name).suffixes == []:
                new_name = f"{new_name}{kwargs['suffix']}"
            if key_path.name != new_name:
                self._clear_storagekey = self.key
                self.key = str(key_path.with_name(new_name))
                logger.warning(f"Replacing the file will also replace the key from '{key_path}' to '{self.key}', and delete '{key_path}' upon `ln.add`")
        else:
            self.key = kwargs["key"]
            old_storage = f"{self.id}{self.suffix}"
            new_storage = self.key if self.key is not None else f"{self.id}{kwargs['suffix']}"
            if old_storage != new_storage:
                self._clear_storagekey = old_storage

        self.suffix = kwargs["suffix"]
        self.size = kwargs["size"]
        self.hash = kwargs["hash"]
        self.run = kwargs["run"]
        self._local_filepath = privates["local_filepath"]
        self._cloud_filepath = privates["cloud_filepath"]
        self._memory_rep = privates["memory_rep"]
        self._to_store = not privates["check_path_in_storage"]  # no need to upload if new file is already in storage

    @property
    def __name__(cls) -> str:
        return "File"

    def __init__(  # type: ignore
        self,
        data: Union[PathLike, DataLike] = None,
        *,
        key: Optional[str] = None,
        name: Optional[str] = None,
        run: Optional[Run] = None,
        format: Optional[str] = None,
        features: List[Features] = None,
        input_of: List[Run] = None,
    ):
        if features is None:
            features = []
        if input_of is None:
            input_of = []
        if not isinstance(features, List):
            features = [features]

        def log_hint(*, check_path_in_storage: bool, key: str, id: str, suffix: str) -> None:
            hint = ""
            if check_path_in_storage:
                hint += "file in storage ✓"
            else:
                hint += "file will be copied to storage upon `ln.add()`"
            if key is None:
                hint += f" using storage key = {id}{suffix}"
            else:
                hint += f" using storage key = {key}"
            logger.hint(hint)

        from lamindb._file import get_file_kwargs_from_data

        kwargs, privates = get_file_kwargs_from_data(
            data=data,
            name=name,
            key=key,
            run=run,
            format=format,
        )
        if features is not None:
            kwargs["features"] = features
        log_hint(
            check_path_in_storage=privates["check_path_in_storage"],
            key=kwargs["key"],
            id=kwargs["id"],
            suffix=kwargs["suffix"],
        )

        # transform cannot be directly passed, just via run
        # it's directly stored in the file table to avoid another join
        # mediate by the run table
        if kwargs["run"] is not None:
            if kwargs["run"].transform_id is not None:
                kwargs["transform_id"] = kwargs["run"].transform_id
                assert kwargs["run"].transform_version is not None
                kwargs["transform_version"] = kwargs["run"].transform_version
            else:
                # accessing the relationship should always be possible if
                # the above if clause was false as then, we should have a fresh
                # Transform object that is not queried from the DB
                assert kwargs["run"].transform is not None
                kwargs["transform"] = kwargs["run"].transform

        super().__init__(**kwargs)
        if data is not None:
            self._local_filepath = privates["local_filepath"]
            self._cloud_filepath = privates["cloud_filepath"]
            self._memory_rep = privates["memory_rep"]
            self._to_store = not privates["check_path_in_storage"]


# add type annotations back asap when re-organizing the module
def storage_key_from_file(file: File):
    if file.key is None:
        return f"{file.id}{file.suffix}"
    else:
        return file.key


# add type annotations back asap when re-organizing the module
def filepath_from_file_or_folder(file_or_folder: Union[File, Folder]):
    from lndb.dev import StorageSettings

    from lndb import settings

    # using __name__ for type check to avoid need of
    # dynamically importing the type
    if file_or_folder.__name__ == "File":
        storage_key = storage_key_from_file(file_or_folder)
    else:
        storage_key = file_or_folder.key
        if storage_key is None:
            raise ValueError("Only real folders have a path!")
    if file_or_folder.storage_id == settings.storage.id:
        path = settings.storage.key_to_filepath(storage_key)
    else:
        logger.warning(
            "file.path() is slow for files outside the currently configured storage"
            " location\nconsider joining for the set of files you're interested in:"
            " ln.select(ln.File, ln.Storage)the path is storage.root / file.key if"
            " file.key is not None\notherwise storage.root / (file.id + file.suffix)"
        )
        import lamindb as ln

        storage = ln.select(ln.Storage, id=file_or_folder.storage_id).one()
        # find a better way than passing None to instance_settings in the future!
        storage_settings = StorageSettings(storage.root, instance_settings=None)
        path = storage_settings.key_to_filepath(storage_key)
    return path



import importlib
from types import ModuleType

import sqlmodel as sqm
from importlib_metadata import requires as importlib_requires
from importlib_metadata import version as get_pip_version
from lamin_logger import logger
from lnhub_rest._assets._schemas import get_schema_module_name
from packaging.requirements import Requirement

from ._db import insert
from ._settings_instance import InstanceSettings
from ._settings_user import UserSettings


def check_schema_version_and_import(schema_name) -> ModuleType:
    def check_version(module_version):
        schema_module_name = get_schema_module_name(schema_name)
        lamindb_version = get_pip_version("lamindb")
        for req in importlib_requires("lamindb"):
            req = Requirement(req)
            if schema_module_name == req.name:
                if not req.specifier.contains(module_version):
                    # it's currently important that we only import lamindb in
                    # case of an error being raised
                    # the following might mask the actual error because this is
                    # raised during instance creation time where lamindb
                    # cannot yet be imported
                    # import lamindb
                    # if lamindb.__version__ != lamindb_version:
                    #     warning = (
                    #         "\nWARNING: importlib_metadata.version('lamindb') gives"
                    #         f" v{lamindb_version}, whereas `import lamindb` gives"
                    #         f" v{lamindb.__version__}"
                    #         f"\nConsider `pip install lamindb=={lamindb_version}`"
                    #     )
                    # else:
                    #     warning = ""
                    raise RuntimeError(
                        f"lamindb v{lamindb_version} needs"
                        f" lnschema_{schema_name}{req.specifier}, you have"
                        f" {module_version}"
                    )

    try:
        # use live version if we can import
        module = importlib.import_module(get_schema_module_name(schema_name))
        module_version = module.__version__
    except Exception as import_error:
        # use pypi version instead
        module_version = get_pip_version(get_schema_module_name(schema_name))
        check_version(module_version)
        raise import_error

    check_version(module_version)
    return module


def load_schema(isettings: InstanceSettings):
    schema_names = ["core"] + list(isettings.schema)
    msg = "Loading schema modules: "
    for schema_name in schema_names:
        module = check_schema_version_and_import(schema_name)
        msg += f"{schema_name}=={module.__version__} "
    return msg, schema_names


def setup_schema(isettings: InstanceSettings, usettings: UserSettings):
    msg, schema_names = load_schema(isettings)
    logger.info(f"{msg}")

    sqm.SQLModel.metadata.create_all(isettings.engine)

    # we could try to also retrieve the user name here at some point
    insert.user(
        email=usettings.email,
        user_id=usettings.id,
        handle=usettings.handle,
        name=usettings.name,
    )

    for schema_name in schema_names:
        schema_module = importlib.import_module(get_schema_module_name(schema_name))
        insert.version(
            schema_module=schema_module,
            user_id=usettings.id,  # type: ignore
            cloud_sqlite=False,
        )
        # this is the only time we need manipulate the migration table
        # in all other cases alembic is going to to do this for us
        schema_id, migration = schema_module._schema_id, schema_module._migration
        if migration is not None:
            table_loc = (
                schema_module.dev if hasattr(schema_module, "dev") else schema_module
            )
            migration_table = getattr(table_loc, f"migration_{schema_id}")
            # we purposefully do not use isettings.session(), here, as we do *not*
            # want to update the local sqlite file from the cloud while looping over
            # schema modules
            # in fact, a synchronization issue led to loss of version information,
            # because the old cloud sqlite file overwrote the newer local file
            with sqm.Session(isettings.engine) as session:
                session.add(migration_table(version_num=migration))
                session.commit()
    isettings._update_cloud_sqlite_file()



"""Link tables.

.. autosummary::
   :toctree: .

   RunInput
   FileFeatures
   ProjectFolder
   FolderFile

"""
from ._link import FileFeatures, FolderFile, ProjectFolder, RunInput  # noqa

# backward compat
RunIn = RunInput



"""SQLModel extensions.

.. autosummary::
   :toctree: .

   BaseORM
"""

import importlib
import re
import typing
from collections import namedtuple
from typing import Any, Iterable, Optional, Sequence, Tuple, Union

import sqlmodel as sqm
from pydantic import create_model
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import relationship as sa_relationship
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.orm.session import object_session
from typeguard import check_type

from .. import __name__

MODULE_NAME = __name__

# add naming convention for alembic
sqm.SQLModel.metadata.naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_`%(constraint_name)s`",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}


def __repr_args__(self) -> Sequence[Tuple[Optional[str], Any]]:
    # sort like fields
    return [(k, self.__dict__[k]) for k in self.__fields__ if (not k.startswith("_sa_") and k in self.__dict__ and self.__dict__[k] is not None)]  # noqa  # noqa


def __repr__(self):
    if object_session(self) is not None:
        return "[session open] " + super(sqm.SQLModel, self).__repr__()
    else:
        return super(sqm.SQLModel, self).__repr__()


sqm.SQLModel.__repr_args__ = __repr_args__
sqm.SQLModel.__repr__ = __repr__
sqm.SQLModel.__str__ = __repr__


class BaseORM(sqm.SQLModel):  # type: ignore
    """SQLModel with lookup, data validation & schema modules."""

    def __init__(self, **user_kwargs):
        # pydantic does not enforce strict type checking for complex types, only their attributes
        validate_relationship_types(self, user_kwargs)
        super().__init__(**user_kwargs)
        validate_with_pydantic(self, user_kwargs)

    @classmethod
    def lookup(cls, field: Optional[str] = None):
        """Lookup rows by field."""
        import lamindb as ln

        if field is None:
            # by default use the name field
            if "name" in cls.__fields__:
                field = "name"
            else:
                non_ids = [i for i in cls.__fields__.keys() if "id" not in i]
                if len(non_ids) > 0:
                    # the first field isn't named with id
                    field = non_ids[0]
                else:
                    # the first field
                    field = next(iter(cls.__fields__.keys()))
        df = ln.select(cls).df()
        values = set(df[field].values)
        keys = _to_lookup_keys(values, padding=cls.__name__)
        return _namedtuple_from_dict(d=dict(zip(keys, values)), name=cls.__name__)

    @declared_attr
    def __tablename__(cls) -> str:  # type: ignore
        """Prefix table name with module name as in Django."""
        return f"{MODULE_NAME}_{cls.__name__.lower()}"


def get_orm(module_name: str):
    global MODULE_NAME
    MODULE_NAME = module_name

    return BaseORM


# backward compat for migrations ---
SCHEMA_NAME = None


def schema_sqlmodel(schema_name: str):
    global SCHEMA_NAME
    SCHEMA_NAME = schema_name

    import lndb

    if lndb.settings.instance.dialect == "sqlite":
        prefix = f"{schema_name}."
        schema_arg = None
        return BaseORM, prefix, schema_arg
    else:
        prefix = ""
        schema_arg = schema_name
        return BaseORM, prefix, schema_arg


def get_sqlite_prefix_schema_delim_from_alembic() -> Tuple[bool, str, Optional[str], str]:
    from alembic import op

    bind = op.get_bind()
    sqlite = bind.engine.name == "sqlite"

    if sqlite:
        prefix, schema, delim = f"{SCHEMA_NAME}.", None, "."
    else:
        prefix, schema, delim = "", SCHEMA_NAME, "_"

    return sqlite, prefix, schema, delim


# end backward compat for migrations ---


def add_relationship_keys(table: sqm.SQLModel):  # type: ignore
    """add all relationship keys to __sqlmodel_relationships__."""
    for i in getattr(table, "__mapper__").relationships:
        if i not in getattr(table, "__sqlmodel_relationships__"):
            getattr(table, "__sqlmodel_relationships__")[i.key] = None


def add_relationship(
    origin_table: sqm.SQLModel,
    origin_attr_name: str,
    target_attr_name: Optional[str] = None,
):
    """Add relationship defined in the origin table to its respective target table.

    Only valid for many-to-many relationships.

    Args:
        origin_table (sqlmodel.SQLModel): origin (input) table
        origin_attr_name (str): name of the attribute in the origin table that defines the relationship
        target_attr_name (str, optional): name of the corresponding attribute to be added in the relationship target table.
            Defaults to either the back_populates parameter in the origin table relationship attribute (if set) or the snake-case, pluralized name of the origin table.

    Returns:
        None

    Raises:
        ValueError: if origin_attr_name does not correspond to a valid relationship attribute in the origin table
        ValueError: if target_attr_name conflicts with the back_populates parameter set in the attribute of the origin table that defines the relationship
    """
    # raise error if there is no attribute with name origin_attr_name
    origin_attr = getattr(origin_table, origin_attr_name)
    # raise error if attr is not a relationship
    relationship = getattr(origin_attr, "prop")
    if not isinstance(relationship, RelationshipProperty):
        raise ValueError("Input attribute must be a relationship.")
    # make sure relationship key is in __sqlmodel_relationships__
    if relationship not in getattr(origin_table, "__sqlmodel_relationships__"):
        getattr(origin_table, "__sqlmodel_relationships__")[relationship.key] = None
    # create relationship attribute in target table
    target_table = relationship.entity.entity
    target_attr_name = _derive_target_attr_name(origin_table, relationship, target_attr_name)
    setattr(
        target_table,
        target_attr_name,  # type: ignore
        sa_relationship(
            origin_table,
            back_populates=relationship.key,
            secondary=relationship.secondary,
        ),
    )
    # add relationship key to target table
    getattr(target_table, "__sqlmodel_relationships__")[target_attr_name] = None


def add_relationships(origin_table: sqm.SQLModel):
    """Add all relationships defined in the origin table to their corresponding target tables.

    Only valid for many-to-many relationships.

    Args:
        origin_table (sqlmodel.SQLModel): origin (input) table

    Returns:
        None
    """
    for attr in origin_table.__dict__.values():
        # skip attributes that are not relationships
        attr_property = getattr(attr, "prop", None)
        if not isinstance(attr_property, RelationshipProperty):
            continue
        relationship = attr_property
        # only configure relationships that are many-to-many
        if getattr(relationship, "secondary") is not None:
            origin_attr_name = relationship.key
            add_relationship(origin_table, origin_attr_name)


def validate_with_pydantic(model, user_kwargs):
    annotations = model.__annotations__
    kwargs = {**model.__dict__, **user_kwargs}
    pydantic_annotations = {}
    for field, ann in annotations.items():
        # ignore special fields
        if field.startswith("_"):
            kwargs.pop(field, None)
            user_kwargs.pop(field, None)
            continue

        # resolve forward references
        resolved_ann = ann
        if "ForwardRef" in str(ann):  # typing.ForwardRef not available in python <= 3.6.12
            resolved_ann = _resolve_forward_ref(ann, model.__module__)

        # handle relationship and optional fields
        if field in model.__sqlmodel_relationships__ or _is_ann_optional(ann) or _is_field_optional(model.__fields__.get(field)):
            pydantic_annotations[field] = (resolved_ann, None)
        else:
            # do not validate auto-populated (server-side) created_at fields
            if field in model.__fields__:
                sa_column_kwargs = getattr(model.__fields__[field].field_info, "sa_column_kwargs", None)
                if isinstance(sa_column_kwargs, dict):
                    if "server_default" in sa_column_kwargs and field not in user_kwargs:
                        continue
            pydantic_annotations[field] = (resolved_ann, ...)

    # validate with pydantic
    validation_model = create_model(model.__class__.__name__, **pydantic_annotations)
    validation_model.validate(kwargs)
    return


def validate_relationship_types(model, user_kwargs):
    # only validate standard sqlmodel relationships (typed) passed by the user
    standard_sqm_relationships = {k: v for k, v in model.__sqlmodel_relationships__.items() if v is not None}
    standard_user_rel_keys = standard_sqm_relationships.keys() & user_kwargs.keys()
    for key in standard_user_rel_keys:
        if "ForwardRef" in str(model.__annotations__[key]):
            rel_type = _resolve_forward_ref(model.__annotations__[key], model.__module__)
        else:
            rel_type = model.__annotations__[key]
        try:
            check_type(key, user_kwargs[key], rel_type)
        except TypeError:
            error_message = (
                f"Validation error for {model.__class__.__name__}\n"
                f"{key}\n"
                f"  should be of type {rel_type.__module__}.{rel_type}, got "
                f"{user_kwargs[key].__class__.__module__}.{user_kwargs[key].__class__.__name__} instead."
            )
            raise TypeError(error_message) from None

    # auto-populate foreign key fields associated with any relationship passed by the user (avoid missing key error during pydantic validation)
    all_user_rel_keys = model.__sqlmodel_relationships__.keys() & user_kwargs.keys()
    for relationship_name in all_user_rel_keys:
        rel_local_fields = getattr(model.__class__, relationship_name).property.local_columns  # model fields associated with the relationship
        rel_fk_fields = [field for field in rel_local_fields if not field.primary_key]  # primary keys are returned as local fields in link table relationships
        rel_fk_field_names = [field.name for field in rel_fk_fields]
        target_col_names = [list(field.foreign_keys)[0].column.name for field in rel_fk_fields]
        for field_name, col_name in zip(rel_fk_field_names, target_col_names):
            if hasattr(model, field_name):
                if getattr(model, field_name) is None:
                    try:
                        setattr(model, field_name, getattr(user_kwargs[relationship_name], col_name))
                    except AttributeError:
                        pass


def _derive_target_attr_name(
    origin_table: sqm.SQLModel,
    origin_relationship: RelationshipProperty,
    custom_target_name,
):
    """Derive the name of the relationship attribute in the target table."""
    # If set, back_populates parameter in the origin relationship takes precedence over any custom naming
    if origin_relationship.back_populates is not None:
        if custom_target_name is not None and custom_target_name != origin_relationship.back_populates:
            raise ValueError(
                "Custom name for the relationship attribute in the target table conflicts with back_populates parameter set in the relationship attribute of the origin table."
            )
        return origin_relationship.back_populates
    else:
        if custom_target_name is not None:
            return custom_target_name
        else:
            # Convert origin table name from camel to snake case and make it plural
            return f"{re.sub(r'(?<!^)(?=[A-Z])', '_', origin_table.__name__).lower()}s"


def _is_ann_optional(type_ann):
    if type_ann.__module__ == "typing":
        try:
            if type_ann.__origin__ == Union:
                if type(None) in type_ann.__args__:
                    return True
        except Exception:
            return False
    return False


def _is_field_optional(model_field):
    if model_field is None:
        return False
    return not model_field.required


def _resolve_forward_ref(ann, module):
    # fetch target type as string
    forward_args = []
    parent_ref = None
    if _is_ann_optional(ann):
        forward_args = [ann.__args__[0].__forward_arg__]
    else:
        try:
            for arg in ann.__args__:
                if "ForwardRef" in str(arg):
                    forward_args += [arg.__forward_arg__]
            parent_ref = ann._name
        except Exception:
            forward_args = [ann.__forward_arg__]

    # fetch target type from string
    resolved_args = []
    for arg in forward_args:
        split_arg = arg.split(".")
        try:
            # if reference is in another module
            resolved_args += [getattr(importlib.import_module(split_arg[0]), split_arg[-1])]
        except Exception:
            resolved_args += [getattr(importlib.import_module(module), split_arg[-1])]

    # return resolved annotation
    if len(resolved_args) == 1:
        if parent_ref is None:
            return resolved_args[0]
        else:
            return getattr(typing, parent_ref)[resolved_args[0]]
    else:
        # This code should never run. SQLmodel is incompatible with non-optional
        # Union-like types (e.g. typing.Union[str, ForwardRef("lnschema_core._core.Transform")])
        # and raises an error at class definition when a field is typed as such.
        return getattr(typing, parent_ref)[resolved_args[0], resolved_args[1]]


# # as a decorator
# def db_lookup(lookup_field: str = 'name'):
#     def wrapper(sqlmodel_class):
#         @classproperty
#         def df(cls):
#             import lamindb as ln

#             return ln.select(sqlmodel_class).df()

#         @classproperty
#         def lookup(cls):
#             return lookup_field

#         sqlmodel_class.df = df
#         sqlmodel_class.lookup = lookup
#         return sqlmodel_class

#     return wrapper


def _to_lookup_keys(x: Iterable[str], padding: str = "LOOKUP") -> list:
    """Convert a list of strings to tab-completion allowed formats."""
    lookup = [re.sub("[^0-9a-zA-Z]+", "_", str(i)) for i in x]
    for i, value in enumerate(lookup):
        if value == "" or (not value[0].isalpha()):
            lookup[i] = f"{padding}_{value}"
    return lookup


def _namedtuple_from_dict(d: dict, name: str = "entity") -> tuple:
    """Create a namedtuple from a dict to allow autocompletion."""
    nt = namedtuple(name, d.keys())  # type:ignore
    return nt(**d)



import os
from pathlib import Path
from typing import Optional, Union

import sqlmodel as sqm
from lamin_logger import logger
from lnhub_rest.core.instance._load_instance import (
    load_instance as load_instance_from_hub,
)

from lndb.dev.upath import UPath

from ._settings import InstanceSettings, settings
from .dev._settings_load import load_instance_settings
from .dev._settings_store import instance_settings_file
from .dev._storage import StorageSettings


def load(
    identifier: str,
    *,
    migrate: Optional[bool] = None,
    storage: Optional[Union[str, Path, UPath]] = None,
    _log_error_message: bool = True,
    _access_token: Optional[str] = None,
) -> Optional[str]:
    """Load existing instance.

    Args:
        identifier: `str` - The instance identifier can the instance name (owner is
            current user), handle/name, or the URL: https://lamin.ai/handle/name.
        storage: `Optional[PathLike] = None` - Load the instance with an
            updated default storage.
        migrate: `Optional[bool] = None` - Whether to auto-migrate or not.

    Returns:
        - "migrate-failed" if migration failed
        - "migrate-success" if migration was successful
        - "migrate-unnecessary" if migration was not required
    """
    owner, name = get_owner_name_from_identifier(identifier)

    hub_result = load_instance_from_hub(
        owner=owner, name=name, _access_token=_access_token
    )
    # if hub_result is not a string, it means it made a request
    # that successfully returned metadata
    if not isinstance(hub_result, str):
        instance_result, storage_result = hub_result
        isettings = InstanceSettings(
            owner=owner,
            name=name,
            storage_root=storage_result.root,
            storage_region=storage_result.region,
            db=instance_result.db,
            schema=instance_result.schema_str,
        )
    else:
        settings_file = instance_settings_file(name, owner)
        if settings_file.exists():
            logger.info(f"Found cached instance metadata: {settings_file}")
            isettings = load_instance_settings(settings_file)
        else:
            if _log_error_message:
                logger.error(
                    f"Instance {owner}/{name} neither loadable from hub nor local"
                    " cache. Check whether instance exists and you have access:"
                    f" https://lamin.ai/{owner}/{name}?tab=collaborators"
                )
            return "instance-not-reachable"

    if storage is not None:
        update_isettings_with_storage(isettings, storage)

    if not isettings.storage.root.exists():
        raise RuntimeError(
            f"Storage root does not exist: {isettings.storage.root}\n"
            "Please amend by passing --storage <my-storage-root>"
        )

    check, msg = isettings._is_db_setup()
    if not check:
        if _log_error_message:
            raise RuntimeError(msg)
        else:
            logger.warning(
                "Instance metadata exists, but DB might have been corrupted or deleted."
                " Re-initializing the DB."
            )
            return "instance-not-reachable"

    message = load_from_isettings(isettings, migrate)
    if not message == "migrate-failed":
        os.environ["LAMINDB_INSTANCE_LOADED"] = "1"
    return message


def get_owner_name_from_identifier(identifier: str):
    if "/" in identifier:
        if identifier.startswith("https://lamin.ai/"):
            identifier = identifier.replace("https://lamin.ai/", "")
        split = identifier.split("/")
        if len(split) > 2:
            raise ValueError(
                "The instance identifier needs to be 'owner/name', the instance name"
                " (owner is current user) or the URL: https://lamin.ai/owner/name."
            )
        owner, name = split
    else:
        owner = settings.user.handle
        name = identifier
    return owner, name


def load_from_isettings(
    isettings: InstanceSettings,
    migrate: Optional[bool] = None,
):
    from ._init_instance import persist_settings_load_schema, register, reload_lamindb
    from ._migrate import check_deploy_migration
    from .dev._setup_knowledge import load_bionty_versions

    persist_settings_load_schema(isettings)
    message = check_deploy_migration(
        usettings=settings.user, isettings=isettings, attempt_deploy=migrate
    )
    if message == "migrate-failed":
        return message
    register(isettings, settings.user)
    load_bionty_versions(isettings)
    reload_lamindb(isettings)
    return message


def update_isettings_with_storage(
    isettings: InstanceSettings, storage: Union[str, Path, UPath]
) -> None:
    isettings._persist()  # this is temporary for import of lnschema_core
    ssettings = StorageSettings(storage, instance_settings=isettings)
    if ssettings.is_cloud:
        try:  # triggering ssettings.id makes a lookup in the storage table
            logger.success(f"Loaded storage: {ssettings.id} / {ssettings.root_as_str}")
        except RuntimeError:
            raise RuntimeError(
                "Storage not registered!\n"
                "Load instance without the `storage` arg and register storage root: "
                f"`lamin set storage --storage {storage}`"
            )
    else:
        # local storage
        # assumption is you want to merely update the storage location
        from lnschema_core import Storage

        isettings._storage = ssettings  # need this here already
        if isettings.dialect == "sqlite":
            isettings._engine = sqm.create_engine(isettings.db)
            with sqm.Session(isettings.engine) as session:
                storage = session.exec(
                    sqm.select(Storage).where(Storage.root == ssettings.root_as_str)
                ).one_or_none()
            if storage is None:
                with sqm.Session(isettings.engine) as session:
                    storage_record = session.exec(sqm.select(Storage)).one()
                    storage_record.root = ssettings.root_as_str
                    session.add(storage_record)
                    session.commit()
                    session.refresh(storage_record)
                logger.success(
                    f"Updated storage root {storage_record.id} to"
                    f" {ssettings.root_as_str}"
                )
        else:
            raise RuntimeError(
                "Cannot currently not update local storage of sqlite upon load. Use"
                " `lamin set --storage`"
            )
    # update isettings in place
    isettings._storage = ssettings



from sqlmodel import Field

from lnschema_core.dev.sqlmodel import get_orm

from . import __name__ as module_name

SQLModel = get_orm(module_name)


class RunInput(SQLModel, table=True):  # type: ignore
    """Inputs of runs.

    This is a many-to-many link table for `run` and `file` storing the
    inputs of data transformations.

    A data transformation can have an arbitrary number of data objects as inputs.

    - The same `file` can be used as input in many different `runs`.
    - One `run` can have several `files` as inputs.
    """

    run_id: str = Field(foreign_key="lnschema_core_run.id", primary_key=True)
    file_id: str = Field(foreign_key="lnschema_core_file.id", primary_key=True)


class FileFeatures(SQLModel, table=True):  # type: ignore
    """Links `File` and `Features`."""

    file_id: str = Field(foreign_key="lnschema_core_file.id", primary_key=True)
    features_id: str = Field(foreign_key="lnschema_core_features.id", primary_key=True)


class ProjectFolder(SQLModel, table=True):  # type: ignore
    """Link table of project and folder."""

    project_id: str = Field(foreign_key="lnschema_core_project.id", primary_key=True)
    folder_id: str = Field(foreign_key="lnschema_core_folder.id", primary_key=True)


class FolderFile(SQLModel, table=True):  # type: ignore
    """Link table of folder and file."""

    folder_id: str = Field(foreign_key="lnschema_core_folder.id", primary_key=True)
    file_id: str = Field(foreign_key="lnschema_core_file.id", primary_key=True)



"""v0.34a2."""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

from lnschema_core.dev.sqlmodel import get_sqlite_prefix_schema_delim_from_alembic

revision = "1c49c9f9a982"
down_revision = "c3f38ffe9e03"


def upgrade() -> None:
    sqlite, prefix, schema, delim = get_sqlite_prefix_schema_delim_from_alembic()

    op.drop_index(f"ix_core{delim}features_created_at", table_name="lnschema_core_features")
    op.drop_index(f"ix_core{delim}features_created_by_id", table_name="lnschema_core_features")
    op.create_index(op.f("ix_lnschema_core_features_created_at"), "lnschema_core_features", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_features_created_by_id"), "lnschema_core_features", ["created_by_id"], unique=False)
    op.drop_index(f"ix_core{delim}file_created_at", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_created_by_id", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_hash", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_key", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_name", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_run_id", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_size", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_storage_id", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_suffix", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_transform_id", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_transform_version", table_name="lnschema_core_file")
    op.drop_index(f"ix_core{delim}file_updated_at", table_name="lnschema_core_file")
    op.create_index(op.f("ix_lnschema_core_file_created_at"), "lnschema_core_file", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_created_by_id"), "lnschema_core_file", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_hash"), "lnschema_core_file", ["hash"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_key"), "lnschema_core_file", ["key"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_name"), "lnschema_core_file", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_run_id"), "lnschema_core_file", ["run_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_size"), "lnschema_core_file", ["size"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_storage_id"), "lnschema_core_file", ["storage_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_suffix"), "lnschema_core_file", ["suffix"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_transform_id"), "lnschema_core_file", ["transform_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_transform_version"), "lnschema_core_file", ["transform_version"], unique=False)
    op.create_index(op.f("ix_lnschema_core_file_updated_at"), "lnschema_core_file", ["updated_at"], unique=False)
    op.drop_index(f"ix_core{delim}folder_created_at", table_name="lnschema_core_folder")
    op.drop_index(f"ix_core{delim}folder_created_by_id", table_name="lnschema_core_folder")
    op.drop_index(f"ix_core{delim}folder_key", table_name="lnschema_core_folder")
    op.drop_index(f"ix_core{delim}folder_name", table_name="lnschema_core_folder")
    op.drop_index(f"ix_core{delim}folder_storage_id", table_name="lnschema_core_folder")
    op.drop_index(f"ix_core{delim}folder_updated_at", table_name="lnschema_core_folder")
    op.create_index(op.f("ix_lnschema_core_folder_created_at"), "lnschema_core_folder", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_folder_created_by_id"), "lnschema_core_folder", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_folder_key"), "lnschema_core_folder", ["key"], unique=False)
    op.create_index(op.f("ix_lnschema_core_folder_name"), "lnschema_core_folder", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_folder_storage_id"), "lnschema_core_folder", ["storage_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_folder_updated_at"), "lnschema_core_folder", ["updated_at"], unique=False)
    op.drop_index(f"ix_core{delim}project_created_at", table_name="lnschema_core_project")
    op.drop_index(f"ix_core{delim}project_created_by_id", table_name="lnschema_core_project")
    op.drop_index(f"ix_core{delim}project_name", table_name="lnschema_core_project")
    op.drop_index(f"ix_core{delim}project_updated_at", table_name="lnschema_core_project")
    op.create_index(op.f("ix_lnschema_core_project_created_at"), "lnschema_core_project", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_project_created_by_id"), "lnschema_core_project", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_project_name"), "lnschema_core_project", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_project_updated_at"), "lnschema_core_project", ["updated_at"], unique=False)
    op.drop_index(f"ix_core{delim}run_created_at", table_name="lnschema_core_run")
    op.drop_index(f"ix_core{delim}run_created_by_id", table_name="lnschema_core_run")
    op.drop_index(f"ix_core{delim}run_external_id", table_name="lnschema_core_run")
    op.drop_index(f"ix_core{delim}run_name", table_name="lnschema_core_run")
    op.drop_index(f"ix_core{delim}run_transform_version", table_name="lnschema_core_run")
    op.create_index(op.f("ix_lnschema_core_run_created_at"), "lnschema_core_run", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_run_created_by_id"), "lnschema_core_run", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_run_external_id"), "lnschema_core_run", ["external_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_run_name"), "lnschema_core_run", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_run_transform_id"), "lnschema_core_run", ["transform_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_run_transform_version"), "lnschema_core_run", ["transform_version"], unique=False)

    op.create_index(op.f("ix_lnschema_core_storage_created_at"), "lnschema_core_storage", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_storage_created_by_id"), "lnschema_core_storage", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_storage_root"), "lnschema_core_storage", ["root"], unique=False)
    op.create_index(op.f("ix_lnschema_core_storage_updated_at"), "lnschema_core_storage", ["updated_at"], unique=False)
    op.drop_index(f"ix_core{delim}transform_created_at", table_name="lnschema_core_transform")
    op.drop_index(f"ix_core{delim}transform_created_by_id", table_name="lnschema_core_transform")
    op.drop_index(f"ix_core{delim}transform_name", table_name="lnschema_core_transform")

    op.create_index(op.f("ix_lnschema_core_transform_created_at"), "lnschema_core_transform", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_created_by_id"), "lnschema_core_transform", ["created_by_id"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_name"), "lnschema_core_transform", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_reference"), "lnschema_core_transform", ["reference"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_title"), "lnschema_core_transform", ["title"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_type"), "lnschema_core_transform", ["type"], unique=False)
    op.create_index(op.f("ix_lnschema_core_transform_updated_at"), "lnschema_core_transform", ["updated_at"], unique=False)
    op.drop_index(f"ix_core{delim}user_created_at", table_name="lnschema_core_user")
    op.drop_index(f"ix_core{delim}user_email", table_name="lnschema_core_user")
    op.drop_index(f"ix_core{delim}user_handle", table_name="lnschema_core_user")
    op.drop_index(f"ix_core{delim}user_name", table_name="lnschema_core_user")
    op.drop_index(f"ix_core{delim}user_updated_at", table_name="lnschema_core_user")
    op.create_index(op.f("ix_lnschema_core_user_created_at"), "lnschema_core_user", ["created_at"], unique=False)
    op.create_index(op.f("ix_lnschema_core_user_email"), "lnschema_core_user", ["email"], unique=True)
    op.create_index(op.f("ix_lnschema_core_user_handle"), "lnschema_core_user", ["handle"], unique=True)
    op.create_index(op.f("ix_lnschema_core_user_name"), "lnschema_core_user", ["name"], unique=False)
    op.create_index(op.f("ix_lnschema_core_user_updated_at"), "lnschema_core_user", ["updated_at"], unique=False)


def downgrade() -> None:
    pass



"""Check and auto-deploy migrations."""

import importlib
import os
from pathlib import Path
from subprocess import run
from types import ModuleType
from typing import Any, Optional

import sqlmodel as sqm
from lamin_logger import logger
from packaging.version import parse as vparse

from lndb._migrate.utils import generate_module_files, modify_alembic_ini
from lndb.dev._db import insert
from lndb.dev._settings_instance import InstanceSettings
from lndb.dev._settings_user import UserSettings
from lndb.dev._setup_schema import get_schema_module_name


def decide_deploy_migration(
    schema_name: str, isettings: InstanceSettings, schema_module: ModuleType
) -> bool:
    schema_id = schema_module._schema_id
    schema_module_name = schema_module.__name__
    current_version = schema_module.__version__
    current_migration = schema_module._migration

    # query the versions table of the schema in the database
    with sqm.Session(isettings.engine) as session:
        table_loc = (
            schema_module.dev if hasattr(schema_module, "dev") else schema_module
        )
        version_table = getattr(table_loc, f"version_{schema_id}")
        result = session.exec(
            sqm.select(version_table.v, version_table.migration).order_by(
                version_table.created_at.desc()
            )
        ).first()
        if result is None:
            logger.warning("No row in the versions table")
            return False
        deployed_version, deployed_migration = result["v"], result["migration"]

    # if there is no migration, yet, we don't need to deploy it  # noqa
    if deployed_migration is None and current_migration is None:
        deploy_migration = False
    # if the current version is smaller than the deployed version  # noqa
    elif vparse(current_version) < vparse(deployed_version):
        # no need to worry if current migration is same as deployed migration
        if current_migration == deployed_migration:
            deploy_migration = False
        else:  # otherwise, raise an error
            raise RuntimeError(
                f"You are trying to connect to an instance ({isettings.identifier})"
                f" that runs v{deployed_version} (migration {deployed_migration}) of"
                f" {schema_module_name} but you only have v{current_version} (migration"
                f" {current_migration}) installed.\nPlease run `pip install"
                f" {schema_module_name}=={deployed_version}`, or install the latest"
                " schema module version from GitHub."
            )
    else:  # if the current version is higher or equal to the latest deployed version
        # no need to worry if current migration is same as deployed migration
        if current_migration == deployed_migration:
            deploy_migration = False
        else:
            logger.warning(
                f"In instance {isettings.identifier}, deployed schema"
                f" {schema_name} v{deployed_version} (migration {deployed_migration})"
                f" is not up to date with installed v{current_version} (migration"
                f" {current_migration})"
            )
            # if migration is confirmed, continue
            if "PYTEST_CURRENT_TEST" not in os.environ:
                logger.warning(
                    "You might need to run the command in a shell if you can't respond"
                    " to the following dialogue"
                )
                response = input(
                    f"Do you want to migrate {schema_name} from"
                    f" {deployed_version}/{deployed_migration} to"
                    f" {current_version}/{current_migration} (y/n)?"
                )
                if response != "y":
                    logger.warning(
                        f"Instance {isettings.identifier} does not match the latest version of schema {schema_name}.\n"  # noqa
                        f"For production use, either install {schema_module_name} {deployed_version} "  # noqa
                        f"or migrate the database to {current_version}."
                    )
                    deploy_migration = False
                else:
                    deploy_migration = True
            else:
                logger.warning(
                    f"Migrate instance {isettings.name} outside a test (CI) run. "
                    "Unexpected errors might happen."
                )
                deploy_migration = False
    return deploy_migration


def check_deploy_migration(
    *,
    usettings: UserSettings,
    isettings: InstanceSettings,
    attempt_deploy: Optional[bool] = None,
):
    # lock the whole migration
    locker = isettings._cloud_sqlite_locker
    locker.lock()
    # synchronize the sqlite file before proceeding
    isettings._update_local_sqlite_file()

    status = []
    schema_names = ["core"] + list(isettings.schema)
    # enforce order (if bionty is part of schema_names, it needs to come 2nd)
    if "bionty" in schema_names:
        schema_names.remove("bionty")
        schema_names.insert(1, "bionty")

    for schema_name in schema_names:
        schema_module_name = get_schema_module_name(schema_name)
        schema_module = importlib.import_module(schema_module_name)
        schema_id = schema_module._schema_id
        current_version = schema_module.__version__

        if attempt_deploy:  # attempt deploy as we want to test migrating the database
            deploy_migration = True
        else:  # check whether we need to deploy based on version comparison
            deploy_migration = decide_deploy_migration(
                schema_name, isettings, schema_module
            )

        if deploy_migration:
            generate_module_files(
                package_name=schema_module_name,
                migrations_path=Path(schema_module.__file__).parent / "migrations",  # type: ignore  # noqa
                schema_id=schema_id,
            )
            migrate_status = deploy(
                version=current_version,
                usettings=usettings,
                isettings=isettings,
                schema_name=schema_name,
                schema_id=schema_id,
                schema_module=schema_module,
            )
            status.append(migrate_status)
        else:
            status.append("migrate-unnecessary")

    locker.unlock()

    if "migrate-failed" in status:
        return "migrate-failed"
    elif "migrate-success" in status:
        return "migrate-success"
    else:
        return "migrate-unnecessary"


def deploy(
    *,
    version: str,
    usettings: UserSettings,
    isettings: InstanceSettings,
    schema_name: str,
    schema_id: str,
    schema_module: Any,
):
    """Migrate database to latest version."""
    schema_root = Path(schema_module.__file__).parent
    filepath = schema_root / "alembic.ini"

    modify_alembic_ini(filepath, isettings, schema_name)

    process = run(
        f"python -m alembic --name {schema_id} upgrade head",
        cwd=f"{schema_root}",
        shell=True,
    )

    if process.returncode == 0:
        logger.success(f"Migrated schema {schema_name} to v{version}")
        # The following call will also update the sqlite file in the cloud.
        insert.version(
            schema_module=schema_module,
            user_id=usettings.id,  # type: ignore
        )
    else:
        logger.error("Automatic migration failed.")

    modify_alembic_ini(filepath, isettings, schema_name, revert=True)

    if process.returncode == 0:
        return "migrate-success"
    else:
        return "migrate-failed"



import os
import sys

import nox
from laminci import move_built_docs_to_docs_slash_project_slug, upload_docs_artifact
from laminci.nox import (
    build_docs,
    login_testuser1,
    login_testuser2,
    run_pre_commit,
    run_pytest,
)

nox.options.default_venv_backend = "none"


LAMIN_ENV = "prod"
if "GITHUB_REF_NAME" in os.environ:
    if os.environ["GITHUB_REF_NAME"] == "main":
        LAMIN_ENV = "prod"
    elif os.environ["GITHUB_REF_NAME"] == "staging":
        LAMIN_ENV = "staging"
elif "LAMIN_ENV" in os.environ:
    LAMIN_ENV = os.environ["LAMIN_ENV"]

env = {"LAMIN_ENV": LAMIN_ENV}


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
def install(session: nox.Session) -> None:
    session.run(*"pip install --no-deps .".split())
    if os.getenv("GITHUB_EVENT_NAME") not in (None, "push"):
        session.run(*"pip install --no-deps ./lnschema-core".split())
    session.run(*"git clone https://github.com/laminlabs/lamindb --depth 1".split())
    if sys.platform.startswith("linux"):  # remove version pin when running on CI
        session.run(*"sed -i /lndb==/d ./lamindb/pyproject.toml".split())
    session.run(*"pip install ./lamindb[bionty,lamin1,aws,test]".split())


@nox.session
def build(session: nox.Session):
    login_testuser1(session, env=env)
    login_testuser2(session, env=env)
    run_pytest(session, env=env)
    build_docs(session)
    upload_docs_artifact()
    move_built_docs_to_docs_slash_project_slug()



from ._settings import settings


def info():
    """Log information about current instance & user."""
    # Accessing cached settings is faster than accessing the hub
    print(settings.user)
    if settings._instance_exists:
        print(settings.instance)
    else:
        print("No instance loaded: lamin load <instance>")



from typing import Optional

from lamin_logger import logger


def track(notebook_path: str, pypackage: Optional[str] = None):
    try:
        from nbproject.dev import initialize_metadata, read_notebook, write_notebook
    except ModuleNotFoundError:
        logger.warning("Install nbproject! pip install nbproject")
        return None

    nb = read_notebook(notebook_path)
    if "nbproject" not in nb.metadata:
        if pypackage is not None:
            pypackage = [pp for pp in pypackage.split(",") if len(pp) > 0]  # type: ignore # noqa
        metadata = initialize_metadata(nb, pypackage=pypackage).dict()
        nb.metadata["nbproject"] = metadata
        write_notebook(nb, notebook_path)
    else:
        logger.info(f"The notebook {notebook_path} is already tracked.")
